import numpy as np
import math
import torch
from tqdm import tqdm
import warnings

SQRTEPS = math.sqrt(float(np.finfo(np.float64).eps))


def compute_ess(logw: np.ndarray):
    r"""
        Compute effective sample size (per centage).

    Parameters
    ----------
    logw : ``np.ndarray``
        Log-weights.
    Returns
    -------
    ess : float
        Effective sample size divided by actual number
        of particles (between 0 and 1)
    """
    logw_max = np.max(logw)
    logw_normed = logw - logw_max

    weights = np.exp(logw_normed) / np.sum(np.exp(logw_normed))
    return 1.0 / np.sum(weights * weights) / len(weights)


def increment_logz(logw: np.ndarray):
    r"""
        Compute log evidence increment factor.

    Parameters
    ----------
    logw : ``np.ndarray``
        Log-weights.
    Returns
    -------
    ess : float
        logZ increment.
    """
    logw_max = np.max(logw)
    logw_normed = logw - logw_max
    N = len(logw)

    return logw_max + np.logaddexp.reduce(logw_normed) - np.log(N)


def resample_equal(samples: np.ndarray,
                   weights: np.ndarray,
                   random_state: int = None):
    """
        Resample a new set of points from the weighted set of inputs
        such that they all have equal weight.
        Each input sample appears in the output array either
        `floor(weights[i] * nsamples)` or `ceil(weights[i] * nsamples)` times,
        with `floor` or `ceil` randomly selected (weighted by proximity).

    Parameters
    ----------
    samples : `~numpy.ndarray` with shape (nsamples,)
        Set of unequally weighted samples.
    weights : `~numpy.ndarray` with shape (nsamples,)
        Corresponding weight of each sample.
    random_state : `int`, optional
        Random seed.

    Returns
    -------
    equal_weight_samples : `~numpy.ndarray` with shape (nsamples,)
        New set of samples with equal weights.
    
    Examples
    --------
    >>> x = np.array([[1., 1.], [2., 2.], [3., 3.], [4., 4.]])
    >>> w = np.array([0.6, 0.2, 0.15, 0.05])
    >>> resample_equal(x, w)
    array([[ 1.,  1.],
           [ 1.,  1.],
           [ 1.,  1.],
           [ 3.,  3.]])
    
    Notes
    -----
    Implements the systematic resampling method described in `Hol, Schon, and
    Gustafsson (2006) <doi:10.1109/NSSPW.2006.4378824>`_.
    """

    if random_state is not None:
        np.random.seed(random_state)

    if abs(np.sum(weights) - 1.) > SQRTEPS:  # same tol as in np.random.choice.
        # Guarantee that the weights will sum to 1.
        warnings.warn("Weights do not sum to 1 and have been renormalized.")
        weights = np.array(weights) / np.sum(weights)

    # Make N subdivisions and choose positions with a consistent random offset.
    n_samples = len(weights)
    positions = (np.random.random() + np.arange(n_samples)) / n_samples

    # Resample the data.
    idx = np.zeros(n_samples, dtype=int)
    cumulative_sum = np.cumsum(weights)
    i, j = 0, 0
    while (i < n_samples) and (j < len(cumulative_sum)):
        if positions[i] < cumulative_sum[j]:
            idx[i] = j
            i += 1
        else:
            j += 1

    return samples[idx]


class ProgressBar:
    """
        Progress bar class.

    Parameters
    ----------
    show : `bool`
        Whether or not to show a progress bar. Default is ``True``.
    """
    def __init__(self, show: bool = True):
        self.progress_bar = tqdm(desc='Iter', disable=not show)
        self.info = dict()

    def update_stats(self, info):
        """
            Update shown stats.

        Parameters
        ----------
        info : dict
            Dictionary with stats to show.
        """
        self.info = {**self.info, **info}
        self.progress_bar.set_postfix(ordered_dict=self.info)

    def update_iter(self):
        """
            Update iteration counter.
        """
        self.progress_bar.update(1)

    def close(self):
        """
            Close progress bar.
        """
        self.progress_bar.close()


class FunctionWrapper(object):
    r"""
        Make the log-likelihood or log-prior function pickleable
        when ``args`` or ``kwargs`` are also included.

    Parameters
    ----------
    f : callable
        Log probability function.
    args : list
        Extra positional arguments to be passed to f.
    kwargs : dict
        Extra keyword arguments to be passed to f.
    """
    def __init__(self, f, args, kwargs):
        self.f = f
        self.args = [] if args is None else args
        self.kwargs = {} if kwargs is None else kwargs

    def __call__(self, x):
        """
            Evaluate log-likelihood or log-prior function.

        Parameters
        ----------
        x : ``np.ndarray``
            Input position array.

        Returns
        -------
        f : float or ``np.ndarray``
            f(x)
        """
        return self.f(x, *self.args, **self.kwargs)


def torch_to_numpy(x: torch.Tensor) -> np.ndarray:
    """
    Cast torch tensor to numpy.

    Parameters
    ----------
    x : torch.Tensor
        Input tensor.

    Returns
    -------
        Numpy array corresponding to the input tensor.
    """
    return x.detach().numpy()


def numpy_to_torch(x: np.ndarray) -> torch.Tensor:
    """
    Cast numpy array to torch tensor.

    Parameters
    ----------
    x : np.ndarray
        Input array.

    Returns
    -------
        Torch tensor corresponding to the input array.
    """
    return torch.tensor(x, dtype=torch.float32)


def torch_double_to_float(x: torch.Tensor, warn: bool = True):
    """
    Cast double precision (Float64) torch tensor to single precision (Float32).

    Parameters
    ----------
    x: torch.Tensor
        Input tensor.
    warn: bool
        If True, warn the user about the typecast.

    Returns
    -------
        Single precision (Float32) torch tensor.
    """
    if x.dtype == torch.float64 and warn:
        warnings.warn(f"Float64 data is currently unsupported, casting to Float32. Output will also have type Float32.")
        return x.float()
    elif x.dtype == torch.float32:
        return x
    else:
        raise ValueError(f"Unsupported datatype for input data: {x.dtype}")



__bibtex__ = """
@article{karamanis2022pocomc,
  title={Accelerating astronomical and cosmological inference with Preconditioned Monte Carlo},
  author={Karamanis, Minas and Beutler, Florian and Peacock, John A and Nabergoj, David, and Seljak, Uro\v{s}},
  journal={in prep},
  year={2022}
}
"""
__url__ = "https://pocomc.readthedocs.io"
__author__ = "Minas Karamanis"
__email__ = "minaskar@gmail.com"
__license__ = "GPL-3.0"
__description__ = "A Python implementation of Preconditioned Monte Carlo for accelerated Bayesian Computation"


from .flow import *
from .sampler import *
from .plotting import *
from ._version import version

__version__ = version



import math
import copy

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.distributions as D


def create_masks(input_size: int,
                 hidden_size: int,
                 n_hidden: int,
                 input_order: str = 'sequential',
                 input_degrees: torch.Tensor = None):
    """
    Helper function to create masks.
    The masks are used to hide certain connections
    in the neural network thus preserving the 
    autoregressive property.
    
    Parameters
    ----------
    input_size : int
        Dimensionality of input data
    hidden_size : int
        Size of hidden layer
    n_hidden : int
        Number of hidden layers
    input_order : str
        Variable order for creating the autoregressive masks: ``"sequential"`` (default) or ``"random"``.
    input_degrees : torch.Tensor or None
        Degrees of connections between layers
    
    Returns
    -------
    masks : torch.Tensor
        Masks
    degrees : torch.Tensor
        Degrees of connections between layers
    """
    # MADE paper sec 4:
    # degrees of connections between layers -- ensure at most in_degree - 1 connections
    degrees = []

    # set input degrees to what is provided in args (the flipped order of the previous layer in a stack of MADEs);
    # else init input degrees based on strategy in input_order (sequential or random)
    if input_order == 'sequential':
        degrees += [torch.arange(input_size)] if input_degrees is None else [input_degrees]
        for _ in range(n_hidden + 1):
            degrees += [torch.arange(hidden_size) % (input_size - 1)]
        degrees += [torch.arange(input_size) % input_size - 1] if input_degrees is None else [
            input_degrees % input_size - 1]

    elif input_order == 'random':
        degrees += [torch.randperm(input_size)] if input_degrees is None else [input_degrees]
        for _ in range(n_hidden + 1):
            min_prev_degree = min(degrees[-1].min().item(), input_size - 1)
            degrees += [torch.randint(min_prev_degree, input_size, (hidden_size,))]
        min_prev_degree = min(degrees[-1].min().item(), input_size - 1)
        degrees += [torch.randint(min_prev_degree, input_size, (input_size,)) - 1] if input_degrees is None else [
            input_degrees - 1]

    # construct masks
    masks = []
    for (d0, d1) in zip(degrees[:-1], degrees[1:]):
        masks += [(d1.unsqueeze(-1) >= d0.unsqueeze(0)).float()]

    return masks, degrees[0]


class MaskedLinear(nn.Linear):
    """
    Masked Linear Layer (MADE building block layer)

    Parameters
    ----------
    input_size : int
        Dimensionality of input data.
    n_outputs : int
        Number of outputs
    mask : torch.Tensor
        Mask used to hide connections.
    cond_label_size : int
        Number of conditional arguments.
    """
    def __init__(self,
                 input_size: int,
                 n_outputs: int,
                 mask: torch.Tensor,
                 cond_label_size: int = None):
        super().__init__(input_size, n_outputs)

        self.register_buffer('mask', mask)

        self.cond_label_size = cond_label_size
        if cond_label_size is not None:
            self.cond_weight = nn.Parameter(torch.rand(n_outputs, cond_label_size) / math.sqrt(cond_label_size))

    def forward(self,
                x: torch.Tensor,
                y: torch.Tensor = None):
        """
        Forward transformation.

        Parameters
        ----------
        x : torch.Tensor
            Input data.
        y : torch.Tensor
            Conditional input data.
        Returns
        -------
        out : torch.Tensor
            Transformed data.
        """
        out = F.linear(x, self.weight * self.mask, self.bias)
        if y is not None:
            out = out + F.linear(y, self.cond_weight)
        return out

    def extra_repr(self):
        return 'in_features={}, out_features={}, bias={}'.format(
            self.in_features, self.out_features, self.bias is not None
        ) + (self.cond_label_size is not None) * ', cond_features={}'.format(self.cond_label_size)


class BatchNorm(nn.Module):
    """
    Batch Normalisation layer

    Parameters
    ----------
    input_size : int
        Dimensionality of input data
    momentum : float
        Value of momentum variable. Default: ``0.9``.
    eps : float
        Value of epsilon parameter. Default: ``1e-5``.
    """
    def __init__(self,
                 input_size: int,
                 momentum: float = 0.9,
                 eps: float = 1e-5):
        super().__init__()
        self.momentum = momentum
        self.eps = eps

        self.batch_mean = None
        self.batch_var = None

        self.log_gamma = nn.Parameter(torch.zeros(input_size))
        self.beta = nn.Parameter(torch.zeros(input_size))

        self.register_buffer('running_mean', torch.zeros(input_size))
        self.register_buffer('running_var', torch.ones(input_size))

    def forward(self,
                x: torch.Tensor,
                cond_y: torch.Tensor = None):
        """
        Forward transformation.
        Parameters
        ----------
        x : torch.Tensor
            Input data.
        cond_y : torch.Tensor
            Conditional input data.
        Returns
        -------
        y : torch.Tensor
            Transformed data.
        log_abs_det_jacobian : torch.Tensor
            log(abs(det(Jacobian)))
        """
        if self.training:
            self.batch_mean = x.mean(0)

            # note MAF paper uses biased variance estimate; ie x.var(0, unbiased=False)
            # If x has a single example, we set the variance to 1.
            self.batch_var = x.var(0) if len(x) > 1 else torch.ones_like(self.batch_mean)

            # update running mean
            self.running_mean.mul_(self.momentum).add_(self.batch_mean.data * (1 - self.momentum))
            self.running_var.mul_(self.momentum).add_(self.batch_var.data * (1 - self.momentum))

            mean = self.batch_mean
            var = self.batch_var
        else:
            mean = self.running_mean
            var = self.running_var

        # compute normalized input (cf original batch norm paper algo 1)
        x_hat = (x - mean) / torch.sqrt(var + self.eps)
        y = self.log_gamma.exp() * x_hat + self.beta

        # compute log_abs_det_jacobian (cf RealNVP paper)
        log_abs_det_jacobian = self.log_gamma - 0.5 * torch.log(var + self.eps)
        return y, log_abs_det_jacobian.expand_as(x)

    def inverse(self,
                y: torch.Tensor,
                cond_y: torch.Tensor = None):
        """
        Inverse transformation.
        Parameters
        ----------
        y : torch.Tensor
            Input data.
        cond_y : torch.Tensor
            Conditional input data.
        Returns
        -------
        x : torch.Tensor
            Transformed data.
        log_abs_det_jacobian : torch.Tensor
            log(abs(det(Jacobian)))
        """
        if self.training:
            mean = self.batch_mean if self.batch_mean is not None else torch.tensor(0.0)
            var = self.batch_var if self.batch_var is not None else torch.tensor(1.0)
        else:
            mean = self.running_mean
            var = self.running_var

        x_hat = (y - self.beta) * torch.exp(-self.log_gamma)
        x = x_hat * torch.sqrt(var + self.eps) + mean

        log_abs_det_jacobian = 0.5 * torch.log(var + self.eps) - self.log_gamma

        return x, log_abs_det_jacobian.expand_as(x)


class FlowSequential(nn.Sequential):
    """ Join multiple layers of a normalizing flow """

    def forward(self, x, y):  # TODO handle signature mismatch
        """
        Forward transformation.
        Parameters
        ----------
        x : torch.Tensor
            Input data.
        y : torch.Tensor
            Conditional input data.
        Returns
        -------
        u : torch.Tensor
            Transformed data.
        log_abs_det_jacobian : torch.Tensor
            log(abs(det(Jacobian)))
        """
        sum_log_abs_det_jacobians = 0
        for module in self:
            x, log_abs_det_jacobian = module(x, y)
            sum_log_abs_det_jacobians = sum_log_abs_det_jacobians + log_abs_det_jacobian
        return x, sum_log_abs_det_jacobians

    def inverse(self,
                u: torch.Tensor,
                y: torch.Tensor):
        """
        Inverse transformation.
        Parameters
        ----------
        u : torch.Tensor
            Input data.
        y : torch.Tensor
            Conditional input data.
        Returns
        -------
        x : torch.Tensor
            Transformed data.
        log_abs_det_jacobian : torch.Tensor
            log(abs(det(Jacobian)))
        """
        sum_log_abs_det_jacobians = 0
        for module in reversed(self):
            u, log_abs_det_jacobian = module.inverse(u, y)
            sum_log_abs_det_jacobians = sum_log_abs_det_jacobians + log_abs_det_jacobian
        return u, sum_log_abs_det_jacobians


class MADE(nn.Module):
    r"""Masked Autoregressive Density Estimator (MADE)

    Parameters
    ----------
    input_size : int
        Dimensionality of inputs
    hidden_size : int
        Size of hidden layers
    n_hidden : int
        Number of hidden layers
    activation : str
        Activation function: ``"relu"`` (default) or ``"tanh"``.
    input_order : str
        Variable order for creating the autoregressive masks:
        ``"sequential"`` (default) or ``"random"``.
    """

    def __init__(self,
                 input_size: int,
                 hidden_size: int,
                 n_hidden: int,
                 cond_label_size=None,
                 activation: str = 'relu',
                 input_order: str = 'sequential',
                 input_degrees=None):
        super().__init__()
        # base distribution for calculation of log prob under the model
        self.register_buffer('base_dist_mean', torch.zeros(input_size))
        self.register_buffer('base_dist_var', torch.ones(input_size))

        # create masks
        masks, self.input_degrees = create_masks(
            input_size,
            hidden_size,
            n_hidden,
            input_order,
            input_degrees
        )

        # setup activation
        if activation == 'relu':
            activation_fn = nn.ReLU()
        elif activation == 'tanh':
            activation_fn = nn.Tanh()
        else:
            raise ValueError('Check activation function.')

        # construct model
        self.net_input = MaskedLinear(input_size, hidden_size, masks[0], cond_label_size)
        self.net = []
        for m in masks[1:-1]:
            self.net += [activation_fn, MaskedLinear(hidden_size, hidden_size, m)]
        self.net += [activation_fn, MaskedLinear(hidden_size, 2 * input_size, masks[-1].repeat(2, 1))]

        self.net = nn.Sequential(*self.net)

    @property
    def base_dist(self):
        """
        Base distribution
        """
        return D.Normal(self.base_dist_mean, self.base_dist_var)

    def forward(self,
                x: torch.Tensor,
                y: torch.Tensor = None):
        """
        Forward transformation.
        Parameters
        ----------
        x : torch.Tensor
            Input data.
        y : torch.Tensor
            Conditional input data.
        Returns
        -------
        Transformed data.
        """
        # MAF eq 4 -- return mean and log std
        m, loga = self.net(self.net_input(x, y)).chunk(chunks=2, dim=1)
        u = (x - m) * torch.exp(-loga)
        # MAF eq 5
        log_abs_det_jacobian = - loga
        return u, log_abs_det_jacobian

    def inverse(self,
                u: torch.Tensor,
                y: torch.Tensor = None,
                sum_log_abs_det_jacobians=None):
        """Inverse transformation.
        
        Parameters
        ----------
        u : torch.Tensor
            Input data.
        y : torch.Tensor
            Conditional input data.
        sum_log_abs_det_jacobians : torch.Tensor
            Sum of the natural logarithm of the jacobian
            determinant of the transformation.

        Returns
        -------
        Transformed data.
        """
        # MAF eq 3
        x = torch.zeros_like(u)
        # run through reverse model
        loga = 0
        for i in self.input_degrees:
            m, loga = self.net(self.net_input(x, y)).chunk(chunks=2, dim=1)
            x[:, i] = u[:, i] * torch.exp(loga[:, i]) + m[:, i]
        log_abs_det_jacobian = loga
        return x, log_abs_det_jacobian

    def log_prob(self,
                 x: torch.Tensor,
                 y: torch.Tensor = None):
        """
        Log-probability of input data
        Parameters
        ----------
        x : torch.Tensor
            Input data.
        y : torch.Tensor
            Conditional input data.
        Returns
        -------
        Log-probability (log-likelihood) of input.
        """
        u, log_abs_det_jacobian = self.forward(x, y)
        return torch.sum(self.base_dist.log_prob(u) + log_abs_det_jacobian, dim=1)


class MAF(nn.Module):
    """
    Masked Autoregressive Flow (MAF).

    Parameters
    ----------
    n_blocks : int
        Number of MADE blocks.
    input_size : int
        Dimensionality of input data.
    hidden_size : int
        Number of neurons per layer.
    n_hidden : int
        Number of layers per MADE block.
    cond_label_size : int
        Dimensionality of conditional input data.
    activation : str
        Activation function: ``"relu"`` (default) or ``"tanh"``.
    input_order : str
        Variable order for creating the autoregressive masks: ``"sequential"`` (default) or ``"random"``.
    batch_norm : bool
        Whether to use batch normalisation (Default is ``True``).
    """

    def __init__(self,
                 n_blocks: int,
                 input_size: int,
                 hidden_size: int,
                 n_hidden: int,
                 cond_label_size: int = None,
                 activation: str = 'relu',
                 input_order: str = 'sequential',
                 batch_norm: bool = True,
                 **kwargs):
        """
        Parameters
        ----------
        n_blocks : int
            Number of MADE blocks.
        input_size : int
            Input dimensionality.
        hidden_size : int
            Hidden dimension in MADE.
        n_hidden : int
            Number of hidden layers in MADE.
        cond_label_size
        activation : str
            Nonlinearity type.
        input_order
        batch_norm : bool
            Use batch normalization.
        kwargs
        """
        
        # TODO write docstring.
        super().__init__()
        # base distribution for calculation of log prob under the model
        self.register_buffer('base_dist_mean', torch.zeros(input_size))
        self.register_buffer('base_dist_var', torch.ones(input_size))

        # construct model
        modules = []
        self.input_degrees = None
        for i in range(n_blocks):
            modules += [
                MADE(input_size, hidden_size, n_hidden, cond_label_size, activation, input_order, self.input_degrees)]
            self.input_degrees = modules[-1].input_degrees.flip(0)
            modules += batch_norm * [BatchNorm(input_size)]

        self.net = FlowSequential(*modules)

    @property
    def base_dist(self):
        """
        Base distribution
        """
        return D.Normal(self.base_dist_mean, self.base_dist_var, validate_args=False)

    def forward(self,
                x: torch.Tensor,
                y: torch.Tensor = None):
        """
        Forward transformation.
        Parameters
        ----------
        x : torch.Tensor
            Input data.
        y : torch.Tensor
            Conditional input data.
        Returns
        -------
        Transformed data.
        """
        return self.net(x, y)

    def inverse(self,
                u: torch.Tensor,
                y: torch.Tensor = None):
        """
        Inverse transformation.
        Parameters
        ----------
        u : torch.Tensor
            Input data.
        y : torch.Tensor
            Conditional input data.
        Returns
        -------
        Transformed data.
        """
        return self.net.inverse(u, y)

    def log_prob(self,
                 x: torch.Tensor,
                 y: torch.Tensor = None):
        """
        Log-probability of input data
        Parameters
        ----------
        x : torch.Tensor
            Input data.
        y : torch.Tensor
            Conditional input data.
        Returns
        -------
        Log-probability (log-likelihood) of input.
        """
        u, sum_log_abs_det_jacobians = self.forward(x, y)
        return torch.sum(self.base_dist.log_prob(u) + sum_log_abs_det_jacobians, dim=1)

    # TODO only implement a single log_prior for both MAF and RealNVP instead of two separate implementations.
    def log_prior(self, scale: float = 1.0, type: str = 'Laplace'):  # TODO rename `type` to avoid builtin shadow
        """
        Log-prior of weights
        Parameters
        ----------
        scale : float
            Scale parameter for prior.
        type : str
            Type of prior to use. Options are ``"Laplace"`` and ``"Gaussian"``.
        Returns
        -------
        Log-prior of weights.
        """
        total = 0.0

        for i, layer in enumerate(self.net):
            if isinstance(layer, MADE):
                for parameter_name, parameter in layer.net.named_parameters():
                    if parameter_name.endswith('weight'):
                        # Regularize weights, but not biases
                        if type in ['Laplace', 'laplace', 'l1', 'L1']:
                            total += parameter.abs().sum()  # Laplace prior
                        elif type in ['Gaussian', 'gaussian', 'l2', 'L2', 'Normal', 'normal']:
                            total += parameter.square().sum()  # Gaussian prior

        if type in ['Laplace', 'laplace', 'l1', 'L1']:
            return -total / scale
        elif type in ['Gaussian', 'gaussian', 'l2', 'L2', 'Normal', 'normal']:
            return -total / (2 * (scale ** 2))


class LinearMaskedCoupling(nn.Module):
    """
    Modified RealNVP Coupling Layers per the MAF paper

    Parameters
    ----------
    input_size : int
        Dimensionality of input data.
    hidden_size : int
        Number of neurons per layer.
    n_hidden : int
        Number of layers per MADE block.
    mask : torch.Tensor
        Mask used to hide connections.
    cond_label_size : int
        Dimensionality of conditional input data.
    """
    def __init__(self,
                 input_size: int,
                 hidden_size: int,
                 n_hidden: int,
                 mask: torch.Tensor,
                 cond_label_size: int = None):
        
        super().__init__()

        self.register_buffer('mask', mask)

        # scale function
        s_net = [nn.Linear(input_size + (cond_label_size if cond_label_size is not None else 0), hidden_size)]
        for _ in range(n_hidden):
            s_net += [nn.Tanh(), nn.Linear(hidden_size, hidden_size)]
        s_net += [nn.Tanh(), nn.Linear(hidden_size, input_size)]
        self.s_net = nn.Sequential(*s_net)

        # translation function
        self.t_net = copy.deepcopy(self.s_net)
        # replace Tanh with ReLU's per MAF paper
        for i in range(len(self.t_net)):
            if not isinstance(self.t_net[i], nn.Linear): self.t_net[i] = nn.ReLU()

    def forward(self,
                x: torch.Tensor,
                y: torch.Tensor = None):
        """
        Forward transformation.
        Parameters
        ----------
        x : torch.Tensor
            Input data.
        y : torch.Tensor
            Conditional input data.
        Returns
        -------
        u : torch.Tensor
            Transformed data.
        log_abs_det_jacobian : torch.Tensor
            log(abs(det(Jacobian)))
        """
        # apply mask
        mx = x * self.mask

        # run through model
        s = self.s_net(mx if y is None else torch.cat([y, mx], dim=1))
        t = self.t_net(mx if y is None else torch.cat([y, mx], dim=1))

        # cf RealNVP eq 8 where u corresponds to x (here we're modeling u)
        u = mx + (1 - self.mask) * (x - t) * torch.exp(-s)

        # log det du/dx; cf RealNVP 8 and 6; note, sum over input_size done at model log_prob
        log_abs_det_jacobian = - (1 - self.mask) * s

        return u, log_abs_det_jacobian

    def inverse(self,
                u: torch.Tensor,
                y: torch.Tensor = None):
        """
        Inverse transformation.
        Parameters
        ----------
        u : torch.Tensor
            Input data.
        y : torch.Tensor
            Conditional input data.
        Returns
        -------
        x : torch.Tensor
            Transformed data.
        log_abs_det_jacobian : torch.Tensor
            log(abs(det(Jacobian)))
        """
        # apply mask
        mu = u * self.mask

        # run through model
        s = self.s_net(mu if y is None else torch.cat([y, mu], dim=1))
        t = self.t_net(mu if y is None else torch.cat([y, mu], dim=1))
        x = mu + (1 - self.mask) * (u * s.exp() + t)  # cf RealNVP eq 7

        log_abs_det_jacobian = (1 - self.mask) * s  # log det dx/du

        return x, log_abs_det_jacobian


class RealNVP(nn.Module):
    r"""
    RealNVP normalising flow.

    Parameters
    ----------
    n_blocks : int
        Number of MADE blocks.
    input_size : int
        Dimensionality of input data.
    hidden_size : int
        Number of neurons per layer.
    n_hidden : int
        Number of layers per MADE block.
    cond_label_size : int
        Dimensionality of conditional input data.
    batch_norm : bool
        Whether to use batch normalisation. Default: ``True``.
    """
    def __init__(self,
                 n_blocks: int,
                 input_size: int,
                 hidden_size: int,
                 n_hidden: int,
                 cond_label_size: int = None,
                 batch_norm: bool = True,
                 **kwargs):
        super().__init__()

        # base distribution for calculation of log prob under the model
        self.register_buffer('base_dist_mean', torch.zeros(input_size))
        self.register_buffer('base_dist_var', torch.ones(input_size))

        # construct model
        modules = []
        mask = torch.arange(input_size).float() % 2
        for i in range(n_blocks):
            modules += [LinearMaskedCoupling(input_size, hidden_size, n_hidden, mask, cond_label_size)]
            mask = 1 - mask
            modules += batch_norm * [BatchNorm(input_size)]

        self.net = FlowSequential(*modules)

    @property
    def base_dist(self):
        """
        Base distribution
        """
        return D.Normal(self.base_dist_mean, self.base_dist_var)

    def forward(self,
                x: torch.Tensor,
                y: torch.Tensor = None):
        """
        Forward transformation.
        Parameters
        ----------
        x : torch.Tensor
            Input data.
        y : torch.Tensor
            Conditional input data.
        Returns
        -------
        Transformed data.
        """
        return self.net(x, y)

    def inverse(self,
                u: torch.Tensor,
                y: torch.Tensor = None):
        """
        Inverse transformation.
        Parameters
        ----------
        u : torch.Tensor
            Input data.
        y : torch.Tensor
            Conditional input data.
        Returns
        -------
        Transformed data.
        """
        return self.net.inverse(u, y)

    def log_prob(self,
                 x: torch.Tensor,
                 y: torch.Tensor = None):
        """
        Log-probability of input data
        Parameters
        ----------
        x : torch.Tensor
            Input data.
        y : torch.Tensor
            Conditional input data.
        Returns
        -------
        Log-probability (log-likelihood) of input.
        """
        u, sum_log_abs_det_jacobians = self.forward(x, y)
        return torch.sum(self.base_dist.log_prob(u) + sum_log_abs_det_jacobians, dim=1)

    def log_prior(self,
                  scale: float = 1.0,
                  type: str = 'Laplace'):  # TODO rename `type` to avoid builtin shadow
        """
        Log-prior of weights
        Parameters
        ----------
        scale : float
            Scale parameter for prior.
        type : str
            Type of prior to use. Options are ``"Laplace"`` and ``"Gaussian"``.
        Returns
        -------
        Log-prior of weights.
        """
        total = 0.0
        for i, layer in enumerate(self.net):
            if isinstance(layer, MADE):
                for parameter_name, parameter in layer.net.named_parameters():
                    if parameter_name.endswith('weight'):
                        # Regularize weights, but not biases
                        if type in ['Laplace', 'laplace', 'l1', 'L1']:
                            total += parameter.abs().sum()  # Laplace prior
                        elif type in ['Gaussian', 'gaussian', 'l2', 'L2', 'Normal', 'normal']:
                            total += parameter.square().sum()  # Gaussian prior

        if type in ['Laplace', 'laplace', 'l1', 'L1']:
            return -total / scale
        elif type in ['Gaussian', 'gaussian', 'l2', 'L2', 'Normal', 'normal']:
            return -total / (2 * (scale ** 2))



"""
Utilities related to logging.
"""
import logging
import os


def setup_logger(output=None, label='pocomc', log_level='WARNING'):
    """
    Setup the logger.
    Based on the implementation in Bilby:
    https://git.ligo.org/lscsoft/bilby/-/blob/master/bilby/core/utils/log.py
    Parameters
    ----------
    output : str, optional
        Path of to output directory.
    label : str, optional
        Label for this instance of the logger.
    log_level : {'ERROR', 'WARNING', 'INFO', 'DEBUG'}, optional
        Level of logging passed to logger.
    Returns
    -------
    :obj:`logging.Logger`
        Instance of the Logger class.
    """
    #from .. import __version__ as version
    if type(log_level) is str:
        try:
            level = getattr(logging, log_level.upper())
        except AttributeError:
            raise ValueError('log_level {} not understood'.format(log_level))
    else:
        level = int(log_level)

    logger = logging.getLogger('pocomc')
    logger.setLevel(level)

    if any([type(h) == logging.StreamHandler for h in logger.handlers]) \
            is False:
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(logging.Formatter(
            '%(asctime)s %(name)s %(levelname)-8s: %(message)s',
            datefmt='%m-%d %H:%M'))
        stream_handler.setLevel(level)
        logger.addHandler(stream_handler)

    if any([type(h) == logging.FileHandler for h in logger.handlers]) is False:
        if label:
            if output:
                if not os.path.exists(output):
                    os.makedirs(output, exist_ok=True)
            else:
                output = '.'
            log_file = os.path.join(output, f'{label}.log')
            file_handler = logging.FileHandler(log_file)
            file_handler.setFormatter(logging.Formatter(
                '%(asctime)s %(levelname)-8s: %(message)s', datefmt='%H:%M'))

            file_handler.setLevel(level)
            logger.addHandler(file_handler)

    for handler in logger.handlers:
        handler.setLevel(level)

    #logger.info(f'Running pocoMC version {version}')
    logger.info(f'Running pocoMC')

    return logger


from typing import Union, Optional, Tuple, Dict, List

from .maf import MAF, RealNVP
import torch
from .tools import torch_double_to_float
import numpy as np
from torch.utils.data import DataLoader, TensorDataset
import copy
import time


def compute_loss(model: Union[MAF, RealNVP],
                 batch: torch.Tensor,
                 device: str,
                 laplace_prior_scale: Optional[float],
                 gaussian_prior_scale: Optional[float],
                 use_context: bool):
    """
    Compute normalising flow loss given a batch of data.
    The loss is defined as the sum of the negative log likelihood and the negative log prior.
    The likelihood is computed according to the normalising flow.
    The prior is the sum of Laplace and Gaussian priors. The prior is unused by default.

    Parameters
    ----------
    model: ``MAF or RealNVP``
        Normalising flow model used to compute the loss.
    batch: ``torch.Tensor``
        Data with shape (n_samples, n_dimensions).
    device: ``str``
        Torch device used for the loss computation.
    laplace_prior_scale: ``float``
        Laplace prior scale.
    gaussian_prior_scale: ``float``
        Gaussian prior scale.
    use_context: ``bool``
        If True, the batch should be a tuple with data as the first element and context data as the second element.
        The likelihood is then further conditional on context data, not only the parameters of the flow.

    Returns
    -------
    Scalar loss value.
    """
    if use_context:
        x = batch[0].to(device)
        y = batch[1].to(device)
        loss = -model.log_prob(x, y).sum()
    else:
        x = batch[0].to(device)
        loss = -model.log_prob(x).sum()

    if laplace_prior_scale is not None:
        loss -= model.log_prior(scale=laplace_prior_scale, type='Laplace')

    if gaussian_prior_scale is not None:
        loss -= model.log_prior(scale=gaussian_prior_scale, type='Gaussian')

    return loss


def fit(model: Union[MAF, RealNVP],
        data: Union[np.ndarray, torch.Tensor],
        context: Union[np.ndarray, torch.Tensor] = None,
        validation_data: Union[np.ndarray, torch.Tensor] = None,
        validation_context: Union[np.ndarray, torch.Tensor] = None,
        validation_split: float = 0.0,
        epochs: int = 20,
        batch_size: int = 100,
        patience: int = np.inf,
        monitor: str = 'val_loss',
        shuffle: bool = True,
        lr: float = 1e-3,
        weight_decay: float = 1e-8,
        clip_grad_norm: float = 1.0,
        laplace_prior_scale: float = 0.2,
        gaussian_prior_scale: float = None,
        device: str = 'cpu',
        verbose: int = 2):
    r"""
    Fit a normalising flow model.
    We minimize the negative log likelihood of data with an optional L1 and/or L2 regularization term.
    The fitting is done using Adam.

    Parameters
    ----------
    model: ``MAF or RealNVP``
        Normalising flow model to fit.
    data : ``np.ndarray``
        Samples used for training the flow.
    context : ``np.ndarray`` or ``None``
        Additional samples corresponding to conditional  arguments. Default: ``None``.
    validation_data : ``np.ndarray`` or ``None``
        Samples used for validating the flow. Default: ``None``.
    validation_context : ``np.ndarray`` or ``None``
        Additional samples corresponding to conditional arguments, used for validation. Default: ``None``.
    validation_split : ``float``
        Percentage of ``data`` to be used for validation. Default: ``0.0``.
    epochs : ``int``
        Number of training epochs. Default: ``20``.
    batch_size : ``int``
        Batch size used for training. Default: ``100``.
    patience : ``int``
        Number of epochs to wait with no improvement in monitored loss until termination.
        Default: ``np.inf`` (never terminate early).
    monitor : ``str``
        Which loss to monitor for early stopping. Must be one of ``val_loss``, ``loss``. Default: ``val_loss``.
    shuffle : ``bool``
        Shuffle data before training. Default: ``True``.
    lr : ``float``
        Learning rate for Adam. Default: ``1e-3``.
    weight_decay : ``float``
        Weight decay for Adam. Default: ``1e-8``.
    clip_grad_norm : ``float``
        Clip large gradients. Default: ``0``.
    laplace_prior_scale : ``float`` or ``None``
        Laplace prior scale for regularisation. Must be nonnegative. Smaller values correspond to more regularization.
        Default: 0.2.
    gaussian_prior_scale : ``float`` or ``None``
        Gaussian prior scale for regularisation. Must be nonnegative. Smaller values correspond to more regularization.
        Default: None (no regularization).
    device : ``str``
        Device to use for training. Default: ``cpu``.
    verbose : ``int``
        Whether to print all (``2``), some (``1``), or no messages (``0``).

    Returns
    -------
    Training history dictionary.
    """

    optimizer = torch.optim.Adam(model.parameters(), lr, weight_decay=weight_decay)

    data = torch.as_tensor(data)
    validation_data = torch.as_tensor(validation_data) if validation_data is not None else None

    if context is not None:
        use_context = True
        context = torch.as_tensor(context)
    else:
        use_context = False

    validation_context = torch.as_tensor(validation_context) if validation_context is not None else None

    if validation_data is not None:
        if use_context:
            train_dl = DataLoader(TensorDataset(data, context), batch_size, shuffle)
            val_dl = DataLoader(TensorDataset(validation_data, validation_context), batch_size, shuffle)
        else:
            train_dl = DataLoader(TensorDataset(data), batch_size, shuffle)
            val_dl = DataLoader(TensorDataset(validation_data), batch_size, shuffle)
        validation = True
    else:
        if 0.0 < validation_split < 1.0:
            validation = True
            split = int(data.size()[0] * (1. - validation_split))
            if use_context:
                data, validation_data = data[:split], data[split:]
                context, validation_context = context[:split], context[split:]
                train_dl = DataLoader(TensorDataset(data, context), batch_size, shuffle)
                val_dl = DataLoader(TensorDataset(validation_data, validation_context), batch_size, shuffle)
            else:
                data, validation_data = data[:split], data[split:]
                train_dl = DataLoader(TensorDataset(data), batch_size, shuffle)
                val_dl = DataLoader(TensorDataset(validation_data), batch_size, shuffle)
        else:
            validation = False
            if use_context:
                train_dl = DataLoader(TensorDataset(data, context), batch_size, shuffle)
            else:
                train_dl = DataLoader(TensorDataset(data), batch_size, shuffle)

    history = dict()  # Collects per-epoch loss
    history['loss'] = []
    history['val_loss'] = []

    if not validation:
        monitor = 'loss'
    best_epoch = 0
    best_loss = np.inf
    best_model = copy.deepcopy(model.state_dict())

    start_time_sec = time.time()

    for epoch in range(epochs):

        # --- TRAIN AND EVALUATE ON TRAINING SET -----------------------------
        model.train()
        train_loss = 0.0

        for batch in train_dl:

            optimizer.zero_grad()
            loss = compute_loss(model, batch, device, laplace_prior_scale, gaussian_prior_scale, use_context)
            loss.backward()
            torch.nn.utils.clip_grad_norm_(model.parameters(), clip_grad_norm)
            optimizer.step()

            # train_loss += loss.data.item() * x.size(0)
            train_loss += loss.data.item()

        train_loss = train_loss / len(train_dl.dataset)

        history['loss'].append(train_loss)

        # --- EVALUATE ON VALIDATION SET -------------------------------------
        model.eval()
        if validation:
            val_loss = 0.0

            for batch in val_dl:
                with torch.no_grad():
                    loss = compute_loss(model, batch, device, laplace_prior_scale, gaussian_prior_scale, use_context)
                # val_loss += loss.data.item() * x.size(0)
                val_loss += loss.data.item()

            val_loss = val_loss / len(val_dl.dataset)

            history['val_loss'].append(val_loss)

        if verbose > 1:
            try:
                print('Epoch %3d/%3d, train loss: %5.2f, val loss: %5.2f' % (epoch + 1, epochs, train_loss, val_loss))
            except:  # TODO specify type of exception
                print('Epoch %3d/%3d, train loss: %5.2f' % (epoch + 1, epochs, train_loss))

        # Monitor loss
        if history[monitor][-1] < best_loss:
            best_loss = history[monitor][-1]
            best_epoch = epoch
            best_model = copy.deepcopy(model.state_dict())

        if epoch - best_epoch >= patience:
            model.load_state_dict(best_model)
            if verbose > 0:
                print('Finished early after %3d epochs' % best_epoch)
                print('Best loss achieved %5.2f' % best_loss)
            break

    # END OF TRAINING LOOP

    if verbose > 0:
        end_time_sec = time.time()
        total_time_sec = end_time_sec - start_time_sec
        time_per_epoch_sec = total_time_sec / epochs
        print()
        print('Time total:     %5.2f sec' % total_time_sec)
        print('Time per epoch: %5.2f sec' % time_per_epoch_sec)

    return history


class Flow:
    r"""
    Normalising Flow class.
    This class implements forward and inverse passes, log density evaluation, sampling and model fitting
    irrespective of the kind of flow used.

    Parameters
    ----------
    ndim : ``int``
        Number of dimensions.
    flow_config : ``dict`` or ``None``
        Configuration of the flow. If ``None`` the default configuration used is ``dict(n_blocks=6,
        hidden_size= 3 * ndim, n_hidden=1, batch_norm=True, activation='relu', input_order='sequential',
        flow_type='maf')``
    train_config : ``dict`` or ``None``
        Training configuration for the flow. If ``None`` the default configuration used is
        ``dict(validation_split=0.2, epochs=1000, batch_size=nparticles, patience=30, monitor='val_loss',
        shuffle=True, lr=[1e-2, 1e-3, 1e-4, 1e-5], weight_decay=1e-8, clip_grad_norm=1.0, laplace_prior_scale=0.2,
        gaussian_prior_scale=None, device='cpu', verbose=0)``
    """
    def __init__(self, ndim: int, flow_config: dict = None, train_config: dict = None):
        
        if ndim == 1:
            raise ValueError(f"1D data is not supported. Please provide data with ndim >= 2.")

        self.ndim = ndim
        self.flow_config = flow_config  # TODO remove this if unused.
        self.train_config = train_config  # TODO remove this if unused.

        self.default_config = dict(
            n_blocks=6,
            hidden_size=3 * self.ndim,
            n_hidden=1,
            batch_norm=True,
            activation='relu',
            input_order='sequential',
            flow_type='maf'
        )
        self.flow = self.create(flow_config)

    def validate_config(self, config: dict):
        """
        Raise a ValueError if the flow config dictionary contains invalid inputs.

        Parameters
        ----------
        config: ``dict``
            dictionary with key-value pairs to be passed to constructors of MAF or RealNVP.
        """
        allowed_keys = list(self.default_config.keys())
        for k in config.keys():
            if k not in allowed_keys:
                raise ValueError(f"Unrecognized config key: {k}. Allowed keys are: {allowed_keys}")

    def create(self, config: dict = None):
        """
        Create a normalizing flow based on the desired flow configuration.

        Parameters
        ----------
        config: ``dict``
            dictionary with key-value pairs to be passed to constructors of MAF or RealNVP.

        Returns
        -------
        NF : ``MAF or RealNVP``
            Normalising flow instance.
        """
        if config is None:
            config = dict()
        self.validate_config(config)
        config = {**self.default_config, **config}  # Overwrite keys in default_config and add new ones

        if config['flow_type'].lower() == 'maf':
            return MAF(input_size=self.ndim, cond_label_size=None, **config)
        elif config['flow_type'].lower() == 'realnvp':
            return RealNVP(input_size=self.ndim, cond_label_size=None, **config)
        else:
            raise ValueError(f"Unsupported flow type: {config['flow_type']}. Please use one of ['maf', 'realnvp'].")

    def fit(self, x: torch.Tensor) -> Dict[str, List]:
        """
        Train the normalising flow on the provided data.

        Parameters
        ----------
        x : ``torch.Tensor``
            Training data with shape (n_samples, n_dimensions).
        
        Returns
        -------
        history: ``dict``
            Training history containing logged scalars from model fitting. Each scalar has its own key and value which
            is a list of per-epoch values.
        """
        x = torch_double_to_float(x)

        default_train_config = dict(
            validation_split=0.2,
            epochs=1000,
            batch_size=x.shape[0],
            patience=30,
            monitor='val_loss',
            shuffle=True,
            lr=[1e-2, 1e-3, 1e-4, 1e-5],
            weight_decay=1e-8,
            clip_grad_norm=1.0,
            laplace_prior_scale=0.2,
            gaussian_prior_scale=None,
            device='cpu',
            verbose=0
        )
        if self.train_config is None:
            self.train_config = dict()
        train_config = {**default_train_config, **self.train_config}

        history = None
        for lr in train_config['lr']:
            train_config_tmp = copy.deepcopy(train_config)
            del train_config_tmp['lr']
            history = fit(self.flow, x, **train_config_tmp, lr=lr)

        return history

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward transformation.
        Inputs are transformed from the original (relating to the distribution to be modeled) to the latent space.

        Parameters
        ----------
        x : ``torch.Tensor``
            Samples to transform.
        Returns
        -------
        u : ``torch.Tensor``
            Transformed samples in latent space with the same shape as the original space inputs.
        """
        x = torch_double_to_float(x)
        return self.flow.forward(x)

    def inverse(self, u: torch.Tensor) -> torch.Tensor:
        """
        Inverse transformation.
        Inputs are transformed from the latent to the original space (relating to the distribution to be modeled).
        
        Parameters
        ----------
        u : ``torch.Tensor``
            Samples to transform.
        Returns
        -------
        x : ``torch.Tensor``
            Transformed samples in the original space with the same shape as the latent space inputs.
        """
        u = torch_double_to_float(u)
        return self.flow.inverse(u)

    def logprob(self, x: torch.Tensor) -> torch.Tensor:  # TODO rename to log_prob
        """
        Compute log probability of samples.
        
        Parameters
        ----------
        x : ``torch.Tensor``
            Input samples
        Returns
        -------
        Log-probability of samples.
        """
        x = torch_double_to_float(x)
        u, log_abs_det_jac = self.flow.forward(x)
        return torch.sum(self.flow.base_dist.log_prob(u) + log_abs_det_jac, dim=1)

    def sample(self, size: int = 1) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Draw random samples from the normalizing flow.

        Parameters
        ----------
        size : ``int``
            Number of samples to generate. Default: 1.
        Returns
        -------
        samples, log_prob : ``tuple``
            Samples as a ``torch.Tensor`` with shape ``(size, n_dimensions)`` and log probability values with shape ``(size, )``.
        """
        u = torch.randn(size, self.ndim)
        x, log_abs_det_jac = self.flow.inverse(u)
        return x, torch.sum(self.flow.base_dist.log_prob(u) + log_abs_det_jac, dim=1)



from typing import List, Union

import numpy as np


def trace(results: dict,
          labels: List = None,
          dims: Union[List, np.ndarray] = None,
          width: float = 10.0,
          height: float = 3.0,
          kde_bins: int = 200):
    r"""Plot particle traces and 1-D marginal posteriors.

    Parameters
    ----------
    results : dict
        Result dictionary produced using ``pocoMC``.
    labels : list or None
        List of parameter names to include in the figure. If ``None``, the labels are set automatically.  Default: None.
    dims : list or `np.ndarray` or None
        The subset of dimensions that should be plotted. If None, all dimensions will be shown. Default: None.
    width : float
        Width of figure. Default: ``10.0``.
    height : float
        Height of each subplot. Default: ``3.0``.
    kde_bins : int
        Number of bins to use for KDE. Default: ``200``.
    
    Returns
    -------
    fig : Figure
        Trace figure.
    """

    import matplotlib.pyplot as plt
    from scipy.stats import gaussian_kde

    # Get samples and log-weights
    samples = results.get("x")
    logw = results.get("logw")
    beta = results.get("beta")

    # Compute weights
    weights = np.exp(logw - np.max(logw, axis=1)[:, np.newaxis])
    weights /= np.sum(weights, axis=1)[:, np.newaxis]

    # Number of beta values, particles and parameters/dimensions
    n_beta, n_particles, n_dim = np.shape(samples)

    # Set labels if None is provided
    if labels is None:
        if dims is None:
            labels = [r"$x_{%s}$" % i for i in range(n_dim)]
        else:
            labels = [r"$x_{%s}$" % i for i in dims]

    if dims is None:
        parameters = np.arange(n_dim)
    else:
        parameters = dims
        n_dim = len(dims)

    # Initialise figure
    fig = plt.figure(figsize=(width, n_dim * height))

    # Loop over parameters/dimensions
    for i, p in enumerate(parameters):
        # Left column -- beta plots
        plt.subplot(n_dim, 2, 2 * i + 1)
        for j in range(n_beta):
            plt.scatter(
                np.full(n_particles, beta[j]),
                samples[j, :, p],
                s=5,
                c='C0',
                alpha=0.1 * weights[j] / np.max(weights[j])
            )
        plt.xscale('log')
        plt.xlabel(r'$\beta$', fontsize=14)
        plt.ylabel(labels[i], fontsize=14)
        plt.xticks(fontsize=12)
        plt.yticks(fontsize=12)

        # Compute 1D KDE for parameter for beta = 1.0
        kde = gaussian_kde(samples[-1, :, p], weights=weights[-1])
        x = np.linspace(np.min(samples[-1, :, p]), np.max(samples[-1, :, p]), kde_bins)

        # Right column -- trace plots
        plt.subplot(n_dim, 2, 2 * i + 2)
        plt.fill_between(x, kde(x), alpha=0.8)
        plt.xlabel(labels[i], fontsize=14)
        plt.ylabel('')
        plt.xticks(fontsize=12)
        plt.yticks([])

    plt.tight_layout()
    return fig


def corner(results: dict,
           labels: List = None,
           dims: Union[List, np.ndarray] = None,
           color: str = None,
           **corner_kwargs):
    r"""Corner plot of the 1-D and 2-D marginal posteriors.

    Parameters
    ----------
    results : dict
        Result dictionary produced using ``pocoMC``.
    labels : list or None
        List of parameter names to include in the figure. If ``labels=None``
        (default) the labels are set automatically.
    dims : list or `np.ndarray` or None
        The subset of dimensions that should be plotted. If not provided, all dimensions will be shown.
    color : str or None
        Color to use for contours.
    **corner_kwargs : -
        Any additional arguments passed to ``corner.corner``.
        See here ``https://corner.readthedocs.io/en/latest/api.html``
        for a complete list.
    
    Returns
    -------
    fig : Figure
        Corner figure.
    """
    import corner

    if color is None:
        color = 'C0'

    # Get posterior samples
    posterior_samples = results.get("samples")

    # Number of  particles and parameters/dimensions
    n_particles, n_dim = np.shape(posterior_samples)

    # Set labels if None is provided
    if labels is None:
        if dims is None:
            labels = [r"$x_{%s}$" % i for i in range(n_dim)]
        else:
            labels = [r"$x_{%s}$" % i for i in dims]

    if dims is not None:
        posterior_samples = posterior_samples[:, dims]

    return corner.corner(data=posterior_samples,
                         labels=labels,
                         color=color,
                         **corner_kwargs)


def run(results: dict,
        full_run: bool = True,
        width: float = 10.0,
        height: float = 10.0):
    r"""Run plot which shows various statistics computed during the run.

    Parameters
    ----------
    results : dict
        Result dictionary produced using ``pocoMC``.
    full_run : bool
        If True, include run diagnostics beyond the basic run. Default: ``True``.
    width : float
        Width of figure. Default: ``10.0``.
    height : float
        Height of each subplot. Default: ``10.0``.
    
    Returns
    -------
    fig : Figure
        Run figure.
    """
    beta = results.get("beta")
    steps = results.get("steps")
    scale = results.get("scale")
    logz = results.get("logz")

    if full_run:
        if len(steps) > len(beta):
            beta = np.hstack((beta, beta[-1] * np.ones(len(steps) - len(beta))))

        if len(steps) > len(logz):
            logz = np.hstack((logz, logz[-1] * np.ones(len(steps) - len(logz))))
    else:
        steps = steps[:len(beta)]
        scale = scale[:len(beta)]

    import matplotlib.pyplot as plt

    # Initialise figure
    fig = plt.figure(figsize=(width, height))

    plt.subplot(4, 1, 1)
    plt.plot(beta, 'o-', lw=2.5)
    plt.ylabel(r"$\beta$", fontsize=16)
    plt.xticks(fontsize=14)
    plt.yticks(fontsize=14)

    plt.subplot(4, 1, 2)
    plt.plot(steps, 'o-', lw=2.5)
    plt.ylabel("steps", fontsize=16)
    plt.xticks(fontsize=14)
    plt.yticks(fontsize=14)

    plt.subplot(4, 1, 3)
    plt.plot(scale, 'o-', lw=2.5)
    plt.ylabel("scale", fontsize=16)
    plt.xticks(fontsize=14)
    plt.yticks(fontsize=14)

    plt.subplot(4, 1, 4)
    plt.plot(logz, 'o-', lw=2.5)
    plt.ylabel(r"$\log \mathcal{Z}$", fontsize=16)
    plt.xlabel("Iterations", fontsize=16)
    plt.xticks(fontsize=14)
    plt.yticks(fontsize=14)

    plt.tight_layout()

    return fig



from typing import Union, List

import numpy as np

from pocomc.input_validation import assert_array_float, assert_array_within_interval


class Reparameterise:
    """
    Class that reparameterises the model using change-of-variables parameter transformations.

    Parameters
    ----------
    ndim : int
        Dimensionality of sampling problem
    bounds : array or list or None
        Parameter bounds
    periodic : list
        List of indices corresponding to parameters with periodic boundary conditions
    reflective : list
        List of indices corresponding to parameters with periodic boundary conditions
    scale : bool
        Rescale parameters to zero mean and unit variance (default is true)
    diagonal : bool
        Use diagonal transformation (i.e. ignore covariance) (default is true)
    """
    def __init__(self,
                 ndim: int,
                 bounds: Union[np.ndarray, list] = None,
                 periodic: List[int] = None,
                 reflective: List[int] = None,
                 scale: bool = True,
                 diagonal: bool = True):

        self.ndim = ndim

        if bounds is None:
            bounds = np.full((self.ndim, 2), np.nan)
        elif len(bounds) == 2 and not np.shape(bounds) == (2, 2):
            bounds = np.tile(np.array(bounds, dtype=np.float32).reshape(2, 1), self.ndim).T
        assert_array_float(bounds)

        self.low = bounds.T[0]
        self.high = bounds.T[1]

        self.periodic = periodic
        self.reflective = reflective

        self.mu = None
        self.sigma = None
        self.cov = None
        self.L = None
        self.L_inv = None
        self.log_det_L = None
        self.scale = scale
        self.diagonal = diagonal

        self._create_masks()

    def apply_boundary_conditions(self, x: np.ndarray):
        """
        Apply boundary conditions (i.e. periodic or reflective) to input.
        The first kind include phase parameters that might be periodic
        e.g. on a range ``[0,2*np.pi]``. The latter can arise in cases
        where parameters are ratios where ``a/b`` and  ``b/a`` are equivalent.

        Parameters
        ----------
        x : np.ndarray
            Input array
        
        Returns
        -------
        Transformed input
        """
        if (self.periodic is None) and (self.reflective is None):
            return x
        elif self.periodic is None:
            return self._apply_reflective_boundary_conditions(x)
        elif self.reflective is None:
            return self._apply_periodic_boundary_conditions(x)
        else:
            return self._apply_reflective_boundary_conditions(self._apply_periodic_boundary_conditions(x))

    def _apply_periodic_boundary_conditions(self, x: np.ndarray):
        """
        Apply periodic boundary conditions to input.
        This can be useful for phase parameters that might be periodic
        e.g. on a range ``[0,2*np.pi]``
        
        Parameters
        ----------
        x : np.ndarray
            Input array
        
        Returns
        -------
        Transformed input.
        """
        if self.periodic is not None:
            x = x.copy()
            for i in self.periodic:
                for j in range(len(x)):
                    while x[j, i] > self.high[i]:
                        x[j, i] = self.low[i] + x[j, i] - self.high[i]
                    while x[j, i] < self.low[i]:
                        x[j, i] = self.high[i] + x[j, i] - self.low[i]
        return x

    def _apply_reflective_boundary_conditions(self, x: np.ndarray):
        """
        Apply reflective boundary conditions to input. This can arise in cases
        where parameters are ratios where ``a/b`` and  ``b/a`` are equivalent.
        
        Parameters
        ----------
        x : np.ndarray
            Input array
        
        Returns
        -------
        Transformed input.
        """
        if self.reflective is not None:
            x = x.copy()
            for i in self.reflective:
                for j in range(len(x)):
                    while x[j, i] > self.high[i]:
                        x[j, i] = self.high[i] - x[j, i] + self.high[i]
                    while x[j, i] < self.low[i]:
                        x[j, i] = self.low[i] + self.low[i] - x[j, i]

        return x

    def fit(self, x: np.ndarray):
        """
        Learn mean and standard deviation using for rescaling.
        
        Parameters
        ----------
        x : np.ndarray
            Input data used for training.
        """
        assert_array_within_interval(x, self.low, self.high)

        u = self._forward(x)
        self.mu = np.mean(u, axis=0)
        if self.diagonal:
            self.sigma = np.std(u, axis=0)
        else:
            self.cov = np.cov(u.T)
            self.L = np.linalg.cholesky(self.cov)
            self.L_inv = np.linalg.inv(self.L)
            self.log_det_L = np.linalg.slogdet(self.L)[1]

    def forward(self, x: np.ndarray):
        """
        Forward transformation (both logit for bounds and affine for all parameters).

        Parameters
        ----------
        x : np.ndarray
            Input data
        Returns
        -------
        u : np.ndarray
            Transformed input data
        """
        assert_array_within_interval(x, self.low, self.high)

        u = self._forward(x)
        if self.scale:
            u = self._forward_affine(u)

        return u

    def inverse(self, u: np.ndarray):
        """
        Inverse transformation (both logit^-1 for bounds and affine for all parameters).

        Parameters
        ----------
        u : np.ndarray
            Input data
        Returns
        -------
        x : np.ndarray
            Transformed input data
        log_det_J : np.array
            Logarithm of determinant of Jacobian matrix transformation.
        """
        if self.scale:
            x, log_det_J = self._inverse_affine(u)
            x, log_det_J_prime = self._inverse(x)
            log_det_J += log_det_J_prime
        else:
            x, log_det_J = self._inverse(u)

        return x, log_det_J

    def _forward(self, x: np.ndarray):
        """
        Forward transformation (only logit for bounds).

        Parameters
        ----------
        x : np.ndarray
            Input data
        Returns
        -------
        u : np.ndarray
            Transformed input data
        """
        u = np.empty(x.shape)
        u[:, self.mask_none] = self._forward_none(x)
        u[:, self.mask_left] = self._forward_left(x)
        u[:, self.mask_right] = self._forward_right(x)
        u[:, self.mask_both] = self._forward_both(x)

        return u

    def _inverse(self, u: np.ndarray):
        """
        Inverse transformation (only logit^-1 for bounds).

        Parameters
        ----------
        u : np.ndarray
            Input data
        Returns
        -------
        x : np.ndarray
            Transformed input data
        log_det_J : np.array
            Logarithm of determinant of Jacobian matrix transformation.
        """
        x = np.empty(u.shape)
        J = np.empty(u.shape)

        x[:, self.mask_none], J[:, self.mask_none] = self._inverse_none(u)
        x[:, self.mask_left], J[:, self.mask_left] = self._inverse_left(u)
        x[:, self.mask_right], J[:, self.mask_right] = self._inverse_right(u)
        x[:, self.mask_both], J[:, self.mask_both] = self._inverse_both(u)

        log_det_J = np.array([np.linalg.slogdet(Ji * np.identity(len(Ji)))[1] for Ji in J])

        return x, log_det_J

    def _forward_affine(self, x: np.ndarray):
        """
        Forward affine transformation.

        Parameters
        ----------
        x : np.ndarray
            Input data
        Returns
        -------
        Transformed input data
        """
        if self.diagonal:
            return (x - self.mu) / self.sigma
        else:
            return np.array([np.dot(self.L_inv, xi - self.mu) for xi in x])

    def _inverse_affine(self, u: np.ndarray):
        """
        Inverse affine transformation.

        Parameters
        ----------
        u : np.ndarray
            Input data
        Returns
        -------
        x : np.ndarray
            Transformed input data
        J : np.ndarray
            Diagonal of Jacobian matrix.
        """
        if self.diagonal:
            J = self.sigma
            log_det_J = np.linalg.slogdet(J * np.identity(len(J)))[1]
            return self.mu + self.sigma * u, log_det_J * np.ones(len(u))
        else:
            x = self.mu + np.array([np.dot(self.L, ui) for ui in u])
            return x, self.log_det_L * np.ones(len(u))

    def _forward_left(self, x: np.ndarray):
        """
        Forward transformation for bounded parameters (only low).

        Parameters
        ----------
        x : np.ndarray
            Input data
        Returns
        -------
        Transformed input data
        """
        return np.log(x[:, self.mask_left] - self.low[self.mask_left])

    def _inverse_left(self, u: np.ndarray):
        """
        Inverse transformation for bounded parameters (only low).

        Parameters
        ----------
        u : np.ndarray
            Input data
        Returns
        -------
        x : np.ndarray
            Transformed input data
        J : np.array
            Diagonal of Jacobian matrix.
        """
        p = np.exp(u[:, self.mask_left])

        return p + self.low[self.mask_left], p

    def _forward_right(self, x: np.ndarray):
        """
        Forward transformation for bounded parameters (only high).

        Parameters
        ----------
        x : np.ndarray
            Input data
        Returns
        -------
        Transformed input data
        """
        return np.log(self.high[self.mask_right] - x[:, self.mask_right])

    def _inverse_right(self, u: np.ndarray):
        """
        Inverse transformation for bounded parameters (only high).

        Parameters
        ----------
        u : np.ndarray
            Input data
        Returns
        -------
        x : np.ndarray
            Transformed input data
        J : np.array
            Diagonal of Jacobian matrix.
        """
        p = np.exp(u[:, self.mask_right])

        return self.high[self.mask_right] - p, p

    def _forward_both(self, x: np.ndarray):
        """
        Forward transformation for bounded parameters (both low and high).

        Parameters
        ----------
        x : np.ndarray
            Input data
        Returns
        -------
        Transformed input data
        """
        p = (x[:, self.mask_both] - self.low[self.mask_both]) / (self.high[self.mask_both] - self.low[self.mask_both])

        return np.log(p / (1 - p))

    def _inverse_both(self, u: np.ndarray):
        """
        Inverse transformation for bounded parameters (both low and high).

        Parameters
        ----------
        u : np.ndarray
            Input data
        Returns
        -------
        x : np.ndarray
            Transformed input data
        J : np.array
            Diagonal of Jacobian matrix.
        """
        p = np.exp(-np.logaddexp(0, -u[:, self.mask_both]))
        x = p * (self.high[self.mask_both] - self.low[self.mask_both]) + self.low[self.mask_both]
        J = (self.high[self.mask_both] - self.low[self.mask_both]) * p * (1.0 - p)
        return x, J

    def _forward_none(self, x:np.ndarray):
        """
        Forward transformation for unbounded parameters (this does nothing).

        Parameters
        ----------
        x : np.ndarray
            Input data
        Returns
        -------
        u : np.ndarray
            Transformed input data
        """
        return x[:, self.mask_none]

    def _inverse_none(self, u:np.ndarray):
        """
        Inverse transformation for unbounded parameters (this does nothing).

        Parameters
        ----------
        u : np.ndarray
            Input data
        Returns
        -------
        x : np.ndarray
            Transformed input data
        log_det_J : np.array
            Logarithm of determinant of Jacobian matrix transformation.
        """
        return u[:, self.mask_none], np.ones(u.shape)[:, self.mask_none]

    def _create_masks(self):
        """
        Create parameter masks for bounded parameters
        """

        self.mask_left = np.zeros(self.ndim, dtype=bool)
        self.mask_right = np.zeros(self.ndim, dtype=bool)
        self.mask_both = np.zeros(self.ndim, dtype=bool)
        self.mask_none = np.zeros(self.ndim, dtype=bool)

        # TODO: Do this more elegantly, it's a shame
        for i in range(self.ndim):
            if np.isnan(self.low[i]) and np.isnan(self.high[i]):
                self.mask_none[i] = True
                self.mask_left[i] = False
                self.mask_right[i] = False
                self.mask_both[i] = False
            elif np.isnan(self.low[i]) and not np.isnan(self.high[i]):
                self.mask_none[i] = False
                self.mask_left[i] = False
                self.mask_right[i] = True
                self.mask_both[i] = False
            elif not np.isnan(self.low[i]) and np.isnan(self.high[i]):
                self.mask_none[i] = False
                self.mask_left[i] = True
                self.mask_right[i] = False
                self.mask_both[i] = False
            else:
                self.mask_none[i] = False
                self.mask_left[i] = False
                self.mask_right[i] = False
                self.mask_both[i] = True



import warnings

import numpy as np
import torch

from .input_validation import assert_array_2d
from .mcmc import preconditioned_metropolis, metropolis
from .tools import resample_equal, FunctionWrapper, numpy_to_torch, compute_ess, increment_logz, ProgressBar
from .scaler import Reparameterise
from .flow import Flow


class Sampler:
    r"""Preconditioned Monte Carlo class.

    Parameters
    ----------
    n_particles : int
        The total number of particles/walkers to use.
    n_dim : int
        The total number of parameters/dimensions.
    log_likelihood : callable
        Function returning the log likelihood of a set
        of parameters.
    log_prior : callable
        Function returning the log prior of a set
        of parameters.
    bounds : ``np.ndarray`` or None
        Array of shape ``(ndim, 2)`` holding the boundaries
        of parameters (default is ``bounds=None``). If a
        parameter is unbounded from below, above or both
        please provide ``None`` for the respective boundary.
    periodic : list
        List of indices that correspond to parameters with
        periodic boundary conditions.
    reflective : list
        List of indices that correspond to parameters with
        reflective boundary conditions.
    threshold : float
        The threshold value for the (normalised) proposal
        scale parameter below which normalising flow
        preconditioning (NFP) is enabled (default is
        ``threshold=1.0``, meaning that NFP is used all
        the time).
    scale : bool
        Whether to scale the distribution of particles to
        have zero mean and unit variance. Default is ``True``.
    rescale : bool
        Whether to rescale the distribution of particles to
        have zero mean and unit variance in every iteration.
        Default is ``False``.
    diagonal : bool
        Use a diagonal covariance matrix when rescaling instead
        of a full covariance. Default is ``True``.
    log_likelihood_args : list
        Extra arguments to be passed to log_likelihood
        (default is ``loglikelihood_args=None``).
    log_likelihood_kwargs : dict
        Extra arguments to be passed to log_likelihood
        (default is ``loglikelihood_kwargs=None``).
    log_prior_args : list
        Extra arguments to be passed to log_prior
        (default is ``log_prior_args=None``).
    log_prior_kwargs : list
        Extra arguments to be passed to log_prior
        (default is ``log_prior_kwargs=None``).
    vectorize_likelihood : bool
        If True, vectorize ``loglikelihood``
        calculation (default is ``vectorize_likelihood=False``).
    vectorize_prior : bool
        If True, vectorize ``log_prior``
        calculation (default is ``vectorize_prior=False``).
    infer_vectorization : bool
        If True, infer the vectorization status of
        the loglikelihood and logprior automatically. Default
        is ``True`` (overwrites the ``vectorize_likelihood``
        and ``vectorize_prior`` arguments).
    pool : pool
        Provided ``MPI`` or ``multiprocessing`` pool for
        parallelisation (default is ``pool=None``).
    parallelize_prior : bool
        If True, use the ``pool`` (if provided)
        for the ``logprior`` as well (default is
        ``parallelize_prior=False``).
    flow_config : dict or ``None``
        Configuration of the normalizing flow (default is
        ``flow_config=None``).
    train_config : dict or ``None``
        Configuration for training the normalizing flow
        (default is ``train_config=None``).
    random_state : int or ``None``
        Initial random seed.
    """
    def __init__(self,
                 n_particles: int,
                 n_dim: int,
                 log_likelihood: callable,
                 log_prior: callable,
                 bounds: np.ndarray = None,
                 periodic=None,
                 reflective=None,
                 threshold: float = 1.0,
                 scale: bool = True,
                 rescale: bool = False,
                 diagonal: bool = True,
                 log_likelihood_args: list = None,
                 log_likelihood_kwargs: dict = None,
                 log_prior_args: list = None,
                 log_prior_kwargs: dict = None,
                 vectorize_likelihood: bool = False,
                 vectorize_prior: bool = False,
                 infer_vectorization: bool = True,
                 pool=None,
                 parallelize_prior: bool = False,
                 flow_config: dict = None,
                 train_config: dict = None,
                 random_state: int = None):
        
        if random_state is not None:
            np.random.seed(random_state)
            torch.manual_seed(random_state)
        self.random_state = random_state

        self.n_walkers = n_particles
        self.n_dim = n_dim

        # Distributions
        self.log_likelihood = FunctionWrapper(
            log_likelihood,
            log_likelihood_args,
            log_likelihood_kwargs
        )
        self.log_prior = FunctionWrapper(
            log_prior,
            log_prior_args,
            log_prior_kwargs
        )

        # Sampling
        self.n_call = 0
        self.t = 0
        self.beta = 0.0
        self.logw = 0.0
        self.sum_logw = 0.0
        self.logz = 0.0

        # Results
        self.saved_iter = []
        self.saved_samples = []
        self.saved_posterior_samples = []
        self.saved_posterior_logl = []
        self.saved_posterior_logp = []
        self.saved_logl = []
        self.saved_logp = []
        self.saved_logw = []
        self.saved_logz = []
        self.saved_ess = []
        self.saved_n_call = []
        self.saved_beta = []
        self.saved_accept = []
        self.saved_scale = []
        self.saved_steps = []

        # State variables
        self.u = None
        self.x = None
        self.J = None
        self.L = None
        self.P = None

        # Parallelism
        self.pool = pool
        if pool is None:
            self.distribute = map
        else:
            self.distribute = pool.map
        self.vectorize_likelihood = vectorize_likelihood
        self.vectorize_prior = vectorize_prior
        self.infer_vectorization = infer_vectorization
        self.parallelize_prior = parallelize_prior

        # Flow
        self.flow = Flow(self.n_dim, flow_config, train_config)
        self.threshold = threshold
        self.use_flow = False

        # Scaler
        self.scaler = Reparameterise(self.n_dim, bounds, periodic, reflective, scale, diagonal)
        self.rescale = rescale

        # MCMC parameters
        self.ideal_scale = 2.38 / np.sqrt(n_dim)
        self.scale = 2.38 / np.sqrt(n_dim)
        self.accept = 0.234
        self.target_accept = 0.234

        # Other
        self.ess = None
        self.gamma = None
        self.n_min = None
        self.n_max = None
        self.progress = None
        self.pbar = None

    def validate_vectorization_settings(self, x_check: np.ndarray):
        """
        Check that vectorization settings are sensible.
        This involves making a likelihood and prior call for each sample in x_check to determine the output shapes.
        User-provided vectorization settings are overwritten according to output shapes.
        The user is warned about the changes.
        If both outputs are arrays (neither is a scalar) of different shape, an error is raised.

        Parameters
        ----------
        x_check: np.ndarray
            Input array used to check vectorization settings.
        """
        def is_function_vectorized(f, n_test=2):
            try:
                output_multiple = f(x_check[:n_test, :])
                if output_multiple.shape == (n_test,):
                    return True
                else:
                    raise ValueError
            except (ValueError, AttributeError):
                output_single = f(x_check[0])
                if isinstance(output_single, float):
                    return False
                else:
                    raise ValueError

        # Use three test samples if ndim = 2
        self.vectorize_likelihood = is_function_vectorized(self.log_likelihood, 3 if self.n_dim == 2 else 2)
        self.vectorize_prior = is_function_vectorized(self.log_prior, 3 if self.n_dim == 2 else 2)

    def run(self,
            prior_samples: np.ndarray,
            ess: float = 0.95,
            gamma: float = 0.75,
            n_min: int = None,
            n_max: int = None,
            progress: bool = True):
        r"""Run Preconditioned Monte Carlo.

        Parameters
        ----------
        prior_samples : ``np.ndarray``
            Array holding the initial positions of the particles. The initial
            positions must be sampled from the prior distribution.
        ess : float
            The effective sample size maintained during the run (default is
            `ess=0.95`).
        gamma : float
            Threshold for the correlation coefficient that is
            used to adaptively determine the number of MCMC
            steps (default is ``gamma=0.75``).
        n_min : int or None
            The minimum number of MCMC steps per iteration (default is ``n_min = ndim // 2``).
        n_max : int or None
            The maximum number of MCMC steps per iteration  (default is ``n_min = int(10 * n_dim)``).
        progress : bool
            If True, print progress bar (default is ``progress=True``).
        """
        assert_array_2d(prior_samples)
        if self.infer_vectorization:
            self.validate_vectorization_settings(prior_samples)  

        # Run parameters
        self.ess = ess
        self.gamma = gamma
        if n_min is None:
            self.n_min = self.n_dim // 2
        else:
            self.n_min = int(n_min)
        if n_max is None:
            self.n_max = int(10 * self.n_dim)
        else:
            self.n_max = int(n_max)
        self.progress = progress

        # Set state parameters
        self.x = np.copy(prior_samples)
        self.scaler.fit(self.x)
        self.u = self.scaler.forward(self.x)
        self.J = self.scaler.inverse(self.u)[1]
        self.P = self._log_prior(self.x)
        self.L = self._log_like(self.x)
        self.n_call += len(self.x)

        # Pre-train flow if required
        if self.threshold >= 1.0:
            self.use_flow = True
            self._train(self.u)

        # Save state
        self.saved_samples.append(self.x)
        self.saved_logl.append(self.L)
        self.saved_logp.append(self.P)
        self.saved_iter.append(self.t)
        self.saved_beta.append(self.beta)
        self.saved_logw.append(np.zeros(self.n_walkers))
        self.saved_logz.append(self.logz)
        self.saved_ess.append(self.ess)
        self.saved_n_call.append(self.n_call)
        self.saved_accept.append(self.accept)
        self.saved_scale.append(self.scale / self.ideal_scale)
        self.saved_steps.append(0)

        # Initialise progress bar
        self.pbar = ProgressBar(self.progress)
        self.pbar.update_stats(
            dict(
                beta=self.beta,
                calls=self.n_call,
                ESS=self.ess,
                logZ=self.logz,
                accept=0.234,
                N=0,
                scale=1.0
            )
        )

        # Run Sequential Monte Carlo
        while 1.0 - self.beta >= 1e-4:

            # Choose next beta based on CV of weights
            self._update_beta()

            # Resample x_prev, w
            self.u, self.x, self.J, self.L, self.P = self._resample(
                self.u,
                self.x,
                self.J,
                self.L,
                self.P
            )

            # Evolve particles using MCMC
            self.u, self.x, self.J, self.L, self.P = self._mutate(
                self.u,
                self.x,
                self.J,
                self.L,
                self.P
            )

            # Rescale parameters
            if self.rescale:
                self.scaler.fit(self.x)
                self.u = self.scaler.forward(self.x)
                self.J = self.scaler.inverse(self.u)[1]

            # Train Preconditioner
            self._train(self.u)

        self.saved_posterior_samples.append(self.x.copy())
        self.saved_posterior_logl.append(self.L.copy())
        self.saved_posterior_logp.append(self.P.copy())

        self.pbar.close()

    def add_samples(self,
                    n: int = 1000,
                    retrain: bool = False,
                    progress: bool = True):
        r"""Method that generates additional samples at the end of the run

        Parameters
        ----------
        n : int
            The number of additional samples. Default: ``1000``.
        retrain : bool
            If True, retrain the normalising flow preconditioner between iterations. Default: ``False``.
        progress : bool
            If True, show progress bar. Default: ``True``.
        """
        self.progress = progress

        self.pbar = ProgressBar(self.progress)
        self.pbar.update_stats(
            dict(
                beta=self.beta,
                calls=self.n_call,
                ESS=self.ess,
                logZ=self.logz,
                accept=0,
                N=0,
                scale=0
            )
        )

        iterations = int(np.ceil(n / len(self.u)))
        for _ in range(iterations):
            self.u, self.x, self.J, self.L, self.P = self._mutate(
                self.u,
                self.x,
                self.J,
                self.L,
                self.P
            )
            if retrain:
                self._train(self.u)
            self.saved_posterior_samples.append(self.x)
            self.saved_posterior_logl.append(self.L)
            self.saved_posterior_logp.append(self.P)
            self.pbar.update_iter()

        self.pbar.close()
        # self.u = np.tile(self.u.T, multiply).T

    def _mutate(self, u, x, J, L, P):
        """
            Method which mutates particle positions

        Parameters
        ----------
        u : ``np.ndarray``
            Scaled positions of particles.
        x : ``np.ndarray``
            Unscaled (original) positions of particles.
        J : ``np.ndarray``
            Logarithms of the absolute determinant of the Jacobian of the scaling transform.
        L : ``np.ndarray``
            log-likelihood values of particles
        P : ``np.ndarray``
            log-prior values of particles

        Returns
        -------
        u : ``np.ndarray``
            Mutated scaled positions of particles.
        x : ``np.ndarray``
            Mutated unscaled (original) positions of particles.
        J : ``np.ndarray``
            Mutated logarithms of the absolute determinant of the Jacobian of the scaling transform.
        L : ``np.ndarray``
            Mutated log-likelihood values of particles
        P : ``np.ndarray``
            Mutated log-prior values of particles
        """
        state_dict = dict(
            u=u.copy(),
            x=x.copy(),
            J=J.copy(),
            L=L.copy(),
            P=P.copy(),
            beta=self.beta
        )

        function_dict = dict(
            loglike=self._log_like,
            logprior=self._log_prior,
            scaler=self.scaler,
            flow=self.flow
        )

        option_dict = dict(
            nmin=self.n_min,
            nmax=self.n_max,
            corr_threshold=self.gamma,
            sigma=self.scale,
            progress_bar=self.pbar
        )

        if self.use_flow:
            results = preconditioned_metropolis(
                state_dict,
                function_dict,
                option_dict
            )
        else:
            results = metropolis(
                state_dict,
                function_dict,
                option_dict
            )

        u = results.get('u').copy()
        x = results.get('x').copy()
        J = results.get('J').copy()
        L = results.get('L').copy()
        P = results.get('P').copy()

        self.scale = results.get('scale')
        n_steps = results.get('steps')
        self.accept = results.get('accept')

        self.n_call += n_steps * len(x)

        self.saved_n_call.append(self.n_call)
        self.saved_accept.append(self.accept)
        self.saved_scale.append(self.scale / self.ideal_scale)
        self.saved_steps.append(n_steps)

        return u, x, J, L, P

    def _train(self, u):
        """
            Method which trains the normalising flow.

        Parameters
        ----------
        u : ``np.ndarray``
            Input training data (i.e. positions of particles)
        """
        if (self.scale < self.threshold * self.ideal_scale and self.t > 1) or self.use_flow:
            y = np.copy(u)
            np.random.shuffle(y)
            self.flow.fit(numpy_to_torch(y))
            self.use_flow = True
        else:
            pass

    def _resample(self, u, x, J, L, P):
        """
            Method which resamples particle positions

        Parameters
        ----------
        u : ``np.ndarray``
            Scaled positions of particles.
        x : ``np.ndarray``
            Unscaled (original) positions of particles.
        J : ``np.ndarray``
            Logarithms of the absolute determinant of the Jacobian of the scaling transform.
        L : ``np.ndarray``
            log-likelihood values of particles
        P : ``np.ndarray``
            log-prior values of particles

        Returns
        -------
        u : ``np.ndarray``
            Resampled scaled positions of particles.
        x : ``np.ndarray``
            Resampled unscaled (original) positions of particles.
        J : ``np.ndarray``
            Resampled logarithms of the absolute determinant of the Jacobian of the scaling transform.
        L : ``np.ndarray``
            Resampled log-likelihood values of particles
        P : ``np.ndarray``
            Resampled log-prior values of particles
        """
        self.saved_samples.append(x)
        self.saved_logl.append(L)
        self.saved_logp.append(P)
        w = np.exp(self.logw - np.max(self.logw))
        w /= np.sum(w)

        assert np.all(~np.isnan(self.logw))
        assert np.all(np.isfinite(self.logw))
        assert np.all(~np.isnan(w))
        assert np.all(np.isfinite(w))

        try:
            idx = resample_equal(np.arange(len(u)), w, self.random_state)
        except IndexError:
            #warnings.warn("Systematic resampling failed. Trying multinomial resampling.")
            idx = np.random.choice(np.arange(len(u)), p=w, size=len(w))
        self.logw = 0.0

        return u[idx], x[idx], J[idx], L[idx], P[idx]

    def _update_beta(self):
        """
            Update beta level and evidence estimate.
        """
        # Update iteration index
        self.t += 1
        self.saved_iter.append(self.t)
        self.pbar.update_iter()

        beta_prev = np.copy(self.beta)
        beta_max = 1.0
        beta_min = np.copy(beta_prev)
        logw_prev = np.copy(self.logw)

        while True:
            beta = (beta_max + beta_min) * 0.5
            self.logw = logw_prev + self.L * (beta - beta_prev)
            ess_est = compute_ess(self.logw)

            if len(self.saved_beta) > 1:
                dbeta = self.saved_beta[-1] - self.saved_beta[-2]

                if 1.0 - beta < dbeta * 0.1:
                    beta = 1.0

            if np.abs(ess_est - self.ess) < min(0.001 * self.ess, 0.001) or beta == 1.0:
                self.saved_beta.append(beta)
                self.saved_logw.append(self.logw)
                self.sum_logw += self.logw
                self.saved_ess.append(ess_est)
                self.beta = beta
                self.pbar.update_stats(dict(beta=self.beta, ESS=ess_est))
                # Update evidence 
                self.logz += increment_logz(self.logw)
                self.saved_logz.append(self.logz)
                self.pbar.update_stats(dict(logZ=self.logz))
                break

            elif ess_est < self.ess:
                beta_max = beta
            else:
                beta_min = beta

    def _log_prior(self, x):
        """
            Compute the log-prior values of the particles.

        Parameters
        ----------
        x : ``np.ndarray``
            Input array of particle positions.

        Returns
        -------
        P : ``np.ndarray``
            Array of log-prior values of particles.
        """
        if self.vectorize_prior:
            return self.log_prior(x)
        elif self.parallelize_prior and self.pool is not None:
            return np.array(list(self.distribute(self.log_prior, x)))
        else:
            return np.array(list(map(self.log_prior, x)))

    def _log_like(self, x):
        """
            Compute the log-likelihood values of the particles.

        Parameters
        ----------
        x : ``np.ndarray``
            Input array of particle positions.

        Returns
        -------
        L : ``np.ndarray``
            Array of log-likelihood values of particles.
        """
        if self.vectorize_likelihood:
            return self.log_likelihood(x)
        elif self.pool is not None:
            return np.array(list(self.distribute(self.log_likelihood, x)))
        else:
            return np.array(list(map(self.log_likelihood, x)))

    def __getstate__(self):
        """
        Get state information for pickling.
        """
        state = self.__dict__.copy()

        try:
            # remove random module
            # del state['rstate']

            # deal with pool
            if state['pool'] is not None:
                del state['pool']  # remove pool
                del state['distribute']  # remove `pool.map` function hook
        except:  # TODO use specific exception type
            pass

        return state

    @property
    def results(self):
        """
            Results dictionary. Includes the following properties: 
            ``samples``, ``loglikelihood``, ``logprior``, ``iter``
            ``logw``, ``logl``, ``logp``, ``logz``, ``ess``, ``ncall``,
            ``beta``, ``accept``, ``scale``, and ``steps``.
        """
        return {
            'samples': np.vstack(self.saved_posterior_samples),
            'loglikelihood': np.hstack(self.saved_posterior_logl),
            'logprior': np.hstack(self.saved_posterior_logp),
            'logz': np.array(self.saved_logz),
            'iter': np.array(self.saved_iter),
            'x': np.array(self.saved_samples),
            'logl': np.array(self.saved_logl),
            'logp': np.array(self.saved_logp),
            'logw': np.array(self.saved_logw),
            'ess': np.array(self.saved_ess),
            'ncall': np.array(self.saved_n_call),
            'beta': np.array(self.saved_beta),
            'accept': np.array(self.saved_accept),
            'scale': np.array(self.saved_scale),
            'steps': np.array(self.saved_steps)
        }



import numpy as np
import torch

from .tools import numpy_to_torch, torch_to_numpy


class Pearson:
    """
    Pearson correlation coefficient.
    This is a measure of the linear correlation between
    two sets of particles. This is used to adaptively
    determine the number of MCMC steps required per iteration
    by comparing the initial particle distribution with
    the current particle distribution and terminating
    MCMC once the correlation coefficient drops below
    a threshold value.

    Parameters
    ----------
    a : ``np.ndarray`` of shape ``(nparticles, ndim)``
        Initial positions of particles
    """

    def __init__(self, a):
        self.l = a.shape[0]  # TODO avoid ambiguous variable name 'l', rename to something else.
        self.am = a - np.sum(a, axis=0) / self.l
        self.aa = np.sum(self.am ** 2, axis=0) ** 0.5

    def get(self, b):
        """
        Method that computes the correlation coefficients between current positions and initial.

        Parameters
        ----------
        b : ``np.ndarray`` of shape ``(n_particles, n_dim)``
            Current positions of particles
        Returns
        -------
        correlation coefficients
        """
        bm = b - np.sum(b, axis=0) / self.l
        bb = np.sum(bm ** 2, axis=0) ** 0.5
        ab = np.sum(self.am * bm, axis=0)
        return np.abs(ab / (self.aa * bb))


@torch.no_grad()
def preconditioned_metropolis(state_dict: dict,
                              function_dict: dict,
                              option_dict: dict):
    """
    Preconditioned Metropolis
    Function that samples the current target distribution
    using a simple Random-walk Metropolis algorithm on the
    latent (i.e. uncorrelated or preconditioned) parameter
    space. The lack of correlations renders Preconditioned
    Metropolis more efficient that standard Metropolis.
    
    Parameters
    ----------
    state_dict : dict
        Dictionary of current state
    function_dict : dict
        Dictionary of functions.
    option_dict : dict
        Dictionary of options.
    
    Returns
    -------
    Results dictionary
    """

    # Clone state variables
    u = torch.clone(numpy_to_torch(state_dict.get('u')))
    x = torch.clone(numpy_to_torch(state_dict.get('x')))
    J = torch.clone(numpy_to_torch(state_dict.get('J')))
    L = torch.clone(numpy_to_torch(state_dict.get('L')))
    P = torch.clone(numpy_to_torch(state_dict.get('P')))
    beta = state_dict.get('beta')
    Z = numpy_to_torch(log_prob(torch_to_numpy(L), torch_to_numpy(P), beta))

    # Get functions
    log_like = function_dict.get('loglike')
    log_prior = function_dict.get('logprior')
    scaler = function_dict.get('scaler')
    flow = function_dict.get('flow')

    # Get MCMC options
    n_min = option_dict.get('nmin')
    n_max = option_dict.get('nmax')
    sigma = option_dict.get('sigma')
    corr_threshold = option_dict.get('corr_threshold')
    progress_bar = option_dict.get('progress_bar')

    # Get number of particles and parameters/dimensions
    n_walkers, n_dim = x.shape

    # Transform u to theta
    theta, log_det_J = flow.forward(u)
    J_flow = -log_det_J.sum(-1)

    # Initialise Pearson correlation object
    corr = Pearson(torch_to_numpy(theta))

    i = 0
    while True:
        i += 1

        # Propose new points in theta space
        theta_prime = theta + sigma * torch.randn(n_walkers, n_dim)

        # Transform to u space
        u_prime, logdetJ_prime = flow.inverse(theta_prime)
        J_flow_prime = logdetJ_prime.sum(-1)

        # Transform to x space
        x_prime, J_prime = scaler.inverse(torch_to_numpy(u_prime))
        x_prime = scaler.apply_boundary_conditions(x_prime)
        x_prime = numpy_to_torch(x_prime)
        J_prime = numpy_to_torch(J_prime)

        # Compute log-likelihood, log-prior, and log-posterior
        L_prime = numpy_to_torch(log_like(torch_to_numpy(x_prime)))
        P_prime = numpy_to_torch(log_prior(torch_to_numpy(x_prime)))
        Z_prime = numpy_to_torch(log_prob(torch_to_numpy(L_prime), torch_to_numpy(P_prime), beta))

        # Compute Metropolis factors
        alpha = torch.minimum(
            torch.ones(n_walkers),
            torch.exp(Z_prime - Z + J_prime - J + J_flow_prime - J_flow)
        )
        alpha[torch.isnan(alpha)] = 0.0

        # Metropolis criterion
        mask = torch.rand(n_walkers) < alpha

        # Accept new points
        theta[mask] = theta_prime[mask]
        u[mask] = u_prime[mask]
        x[mask] = x_prime[mask]
        J[mask] = J_prime[mask]
        J_flow[mask] = J_flow_prime[mask]
        Z[mask] = Z_prime[mask]
        L[mask] = L_prime[mask]
        P[mask] = P_prime[mask]

        # Adapt scale parameter using diminishing adaptation
        sigma_prime = sigma + 1 / (i + 1) * (torch.mean(alpha) - 0.234)
        if sigma_prime > 1e-4:
            sigma = sigma_prime

        # Compute correlations
        cc_prime = corr.get(torch_to_numpy(theta))

        # Update progress bar if available
        if progress_bar is not None:
            progress_bar.update_stats(
                dict(
                    calls=progress_bar.info['calls'] + n_walkers,
                    accept=torch.mean(alpha).item(),
                    N=i,
                    scale=sigma.item() / (2.38 / np.sqrt(n_dim)),
                    corr=np.mean(cc_prime)
                )
            )

        # Loop termination criteria:
        if corr_threshold is None and i >= int(n_min * ((2.38 / np.sqrt(n_dim)) / sigma.item()) ** 2):
            break
        elif np.mean(cc_prime) < corr_threshold and i >= n_min:
            break
        elif i >= n_max:
            break

    return dict(
        u=torch_to_numpy(u),
        x=torch_to_numpy(x),
        J=torch_to_numpy(J),
        L=torch_to_numpy(L),
        P=torch_to_numpy(P),
        scale=sigma.item(),
        accept=torch.mean(alpha).item(),
        steps=i
    )


def metropolis(state_dict: dict,
               function_dict: dict,
               option_dict: dict):
    """
    Random-walk Metropolis
    Function that samples the current target distribution
    using a simple Random-walk Metropolis algorithm (i.e. 
    Metropolis-Hastings with Normal proposal distribution).
    
    Parameters
    ----------
    state_dict : dict
        Dictionary of current state
    function_dict : dict
        Dictionary of functions.
    option_dict : dict
        Dictionary of options.
    
    Returns
    -------
    Results dictionary
    """

    # Clone state variables
    u = state_dict.get('u').copy()
    x = state_dict.get('x').copy()
    J = state_dict.get('J').copy()
    L = state_dict.get('L').copy()
    P = state_dict.get('P').copy()
    beta = state_dict.get('beta')
    Z = log_prob(L, P, beta)

    # Get functions
    log_like = function_dict.get('loglike')
    log_prior = function_dict.get('logprior')
    scaler = function_dict.get('scaler')

    # Get MCMC options
    n_min = option_dict.get('nmin')
    n_max = option_dict.get('nmax')
    sigma = option_dict.get('sigma')
    corr_threshold = option_dict.get('corr_threshold')
    progress_bar = option_dict.get('progress_bar')

    # Get number of particles and parameters/dimensions
    n_walkers, n_dim = x.shape

    # Compute proposal sample covariance and lower triangular Cholesky in u-space
    cov = np.cov(u.T)
    L_triangular = np.linalg.cholesky(cov)

    # Initialise Pearson correlation object
    corr = Pearson(u)

    i = 0
    while True:
        i += 1

        # Propose new points in u space
        u_prime = u + sigma * np.dot(L_triangular, np.random.randn(n_walkers, n_dim).T).T

        # Transform to x space
        x_prime, J_prime = scaler.inverse(u_prime)
        x_prime = scaler.apply_boundary_conditions(x_prime)

        # Compute log-likelihood, log-prior, and log-posterior
        L_prime = log_like(x_prime)
        P_prime = log_prior(x_prime)
        Z_prime = log_prob(L_prime, P_prime, beta)

        # Compute Metropolis factor
        alpha = np.minimum(
            np.ones(len(u_prime)),
            np.exp(Z_prime - Z + J_prime - J)
        )
        alpha[np.isnan(alpha)] = 0.0

        # Metropolis criterion
        mask = np.random.rand(n_walkers) < alpha

        # Accept new points
        u[mask] = u_prime[mask]
        x[mask] = x_prime[mask]
        J[mask] = J_prime[mask]
        Z[mask] = Z_prime[mask]
        L[mask] = L_prime[mask]
        P[mask] = P_prime[mask]

        # Adapt scale parameter using diminishing adaptation
        sigma_prime = sigma + 1 / (i + 1) * (np.mean(alpha) - 0.234)
        if sigma_prime > 1e-4:
            sigma = sigma_prime

        # Compute correlation coefficient
        cc_prime = corr.get(u)

        # Update progress bar if available
        if progress_bar is not None:
            progress_bar.update_stats(
                dict(
                    calls=progress_bar.info['calls'] + len(u),
                    accept=np.mean(alpha),
                    N=i,
                    scale=sigma / (2.38 / np.sqrt(n_dim)),
                    corr=np.mean(cc_prime),
                )
            )

        # Termination criteria:
        if corr_threshold is None and i >= int(n_min * ((2.38 / np.sqrt(n_dim)) / sigma) ** 2):
            break
        elif np.mean(cc_prime) < corr_threshold and i >= n_min:
            break
        elif i >= n_max:
            break

    return dict(
        u=u,
        x=x,
        J=J,
        L=L,
        P=P,
        scale=sigma,
        accept=np.mean(alpha),
        steps=i
    )


def log_prob(L, P, beta):
    """
    Helper function that computes tempered log posterior.
    
    Parameters
    ----------
    L : ``np.ndarray``
        Log-likelihood array
    P : ``np.ndarray``
        Log-prior array
    beta : ``float``
        Beta value
    
    Returns
    -------
    Log-posterior array
    """

    L[np.isnan(L)] = -np.inf
    L[np.isnan(P)] = -np.inf
    L[~np.isfinite(P)] = -np.inf
    P[np.isnan(P)] = -np.inf

    return P + beta * L
